/*
 * PROGRAM:
 * vtkcellgen
 * 
 * PURPOSE:
 * vtkcellgen is a very simple cell generator to supports the generation
 * of several linear, quadratic and cubic VTK cell types.
 * Based on one cell type a flat plate (unstructured grid) will be created with a specific number of 
 * cells in x and y direction. The output is written as
 * VTK legacy ASCII format to stdout or a specified file.
 *
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 *
 * DATE:
 * Berlin, 11 Jun 2007
 *
 * VERSION:
 * 0.1
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

/*default grid size*/
#define CELLNUMX 1
#define CELLNUMY 1
#define DX 1.0
#define DY 1.0
#define DZ 1.0
#define DEFAULT_CELL VTK_QUAD


/*available cell types*/
#define  VTK_POLY_VERTEX		       2
#define  VTK_QUAD                              9
#define  VTK_HEXAHEDRON			       12
#define  VTK_QUADRATIC_TRIANGLE                22
#define  VTK_QUADRATIC_QUAD                    23
#define  VTK_QUADRATIC_HEXAHEDRON              25
#define  VTK_QUADRATIC_WEDGE                   26
#define  VTK_BIQUADRATIC_QUAD                  28
#define  VTK_TRIQUADRATIC_HEXAHEDRON           29
#define  VTK_QUADRATIC_LINEAR_QUAD             30
#define  VTK_QUADRATIC_LINEAR_WEDGE            31
#define  VTK_BIQUADRATIC_QUADRATIC_WEDGE       32
#define  VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON  33
#define  VTK_CUBIC_EDGE                        34
#define  VTK_BICUBIC_QUAD                      36
#define  VTK_TRICUBIC_HEXAHEDRON               37

/*local prototypes*/
static int  get_number_of_points_of_cell (int celltype);
static void print_ugrid_header (FILE * out);
static void print_ugrid_cells (FILE * out, int cellx, int celly, int celltype);
static void print_ugrid_points (FILE * out, int xnum, int ynum, int znum, double dx, double dy, double dz, int zelev);
static void print_ugrid_point_data (FILE * out, int xnum, int ynum, int znum, int pdata);
static void compute_node_numbers (int cellx, int celly, int celltype, int *xnum, int *ynum, int *znum);
static int  print_single_cell (FILE * out, int cols, int rows, int depths, int *xoffste, int *yoffset, int *zoffset,
			      int celltype);
static char *parse_parameter (int argc, char **argv, int *celltype, int *cellx, int *celly, double *dx, double *dy,
			      double *dz, int *zelev, int *pdata);
static void print_help (void);

static int  quite = 0;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int
main (int argc, char **argv)
{
  int celltype;
  int cellx, celly;
  char *filename = NULL;
  FILE *out;
  int xnum, ynum, znum;
  double dx, dy, dz;
  int zelev, pdata = 0;

  /*parse the arguments and set the parameter */
  filename = parse_parameter (argc, argv, &celltype, &cellx, &celly, &dx, &dy, &dz, &zelev, &pdata);

  /*open file */
  if (filename != NULL)
    {
      if (!(out = fopen (filename, "w+")))
	{
	  fprintf (stderr, "Unable to open file %s\n", filename);
	  exit (-1);
	}
    }
  else
    {
      out = stdout;
    }

  /*compute the number of needed nodes */
  compute_node_numbers (cellx, celly, celltype, &xnum, &ynum, &znum);
  /*the header */
  print_ugrid_header (out);
  /*points */
  print_ugrid_points (out, xnum, ynum, znum, dx, dy, dz, zelev);
  /*cells */
  print_ugrid_cells (out, cellx, celly, celltype);
  /*point data */
  print_ugrid_point_data (out, xnum, ynum, znum, pdata);

  return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_help (void)
{
  printf ("vtkcellgen [OPTIONS] \n");
  printf ("Generate a plate (unstructured grid) of a specific VTK cell\n");
  printf ("\n");
  printf ("vtkcellgen is a very simple cell generator to support the generation\n");
  printf ("of several linear, quadratic and cubic VTK cell types.\n");
  printf ("Based on one cell type a flat plate (unstructured grid) will be created with a specific number of \n");
  printf ("cells in x and y direction. The output is written as \n");
  printf ("VTK legacy ASCII format to stdout or a specified file.\n");
  printf ("\n");
  printf ("The following options are supported:          \n");
  printf ("  --help, -h, -help, help      show this help and exit\n");
  printf ("  -z                           vary the z elevation with a sin/cos function\n");
  printf ("  -p                           use point indices as point data, as default a sin/cos function is used\n");
  printf ("                               to generate point data\n");
  printf ("  -q                           quite\n");
  printf ("  --celltype=VTK_CELL_TYPE     set the cell type which should\n");
  printf ("                               be generated, default: %i\n", DEFAULT_CELL);
  printf ("  --cellx=integer              the number of cells in x direction, default: %i\n", CELLNUMX);
  printf ("  --celly=integer              the number of cells in y direction, default: %i\n", CELLNUMY);
  printf ("  --dx=double                  point distance in x direction, default: %g\n", DX);
  printf ("  --dy=double                  point distance in y direction, default: %g\n", DY);
  printf ("  --dz=double                  point distance in z direction, default: %g\n", DZ);
  printf ("  --output=filename            the name of the output file, eg: quads.vtk\n");
  printf ("                               default: stdout\n");
  printf ("\n");
  printf ("The following cell types are supported:     \n");
  printf ("                                            \n");
  printf ("    VTK_POLY_VERTEX                       2 \n");
  printf ("    VTK_QUAD                              9 \n");
  printf ("    VTK_HEXAHEDRON                        12\n");
  printf ("    VTK_QUADRATIC_TRIANGLE                22\n");
  printf ("    VTK_QUADRATIC_QUAD                    23\n");
  printf ("    VTK_QUADRATIC_HEXAHEDRON              25\n");
  printf ("    VTK_QUADRATIC_WEDGE                   26\n");
  printf ("    VTK_BIQUADRATIC_QUAD                  28\n");
  printf ("    VTK_TRIQUADRATIC_HEXAHEDRON           29\n");
  printf ("    VTK_QUADRATIC_LINEAR_QUAD             30\n");
  printf ("    VTK_QUADRATIC_LINEAR_WEDGE            31\n");
  printf ("    VTK_BIQUADRATIC_QUADRATIC_WEDGE       32\n");
  printf ("    VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON  33\n");
  printf ("    VTK_CUBIC_EDGE                        34\n");
  printf ("    VTK_BICUBIC_QUAD                      36\n");
  printf ("    VTK_TRICUBIC_HEXAHEDRON               37\n");
  printf ("\n");
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
char *
parse_parameter (int argc, char **argv, int *celltype, int *cellx, int *celly,
		 double *dx, double *dy, double *dz, int *zelev, int *pdata)
{
  int i, j, stringlength;
  char *buff;
  char *token;
  char *filename = NULL;

  /*default values */
  *celltype = DEFAULT_CELL;
  *cellx = CELLNUMX;
  *celly = CELLNUMY;
  *dx = DX;
  *dy = DY;
  *dz = DZ;
  *zelev = 0;
  *pdata = 0;
  quite = 0;

  /*parse the arguments */
  for (i = 0; i < argc; i++)
    {
      stringlength = strlen (argv[i]);
      if (strncmp (argv[i], "--", 2) == 0)
	{

	  buff = (char *) calloc (stringlength, sizeof (char));
	  token = (char *) calloc (stringlength, sizeof (char));

	  for (j = 0; j < stringlength; j++)
	    {
	      buff[j] = 0;
	      token[j] = 0;
	    }

	  /*copy string */
	  for (j = 2; j < stringlength; j++)
	    buff[j - 2] = argv[i][j];

	  strcat (buff, "\0");

	  if (strncmp (buff, "cellx=", 6) == 0)
	    {
	      for (j = 6; j < stringlength; j++)
		token[j - 6] = buff[j];

	      sscanf (token, "%i", cellx);
	    }

	  if (strncmp (buff, "celly=", 6) == 0)
	    {
	      for (j = 6; j < stringlength; j++)
		token[j - 6] = buff[j];

	      sscanf (token, "%i", celly);
	    }

	  if (strncmp (buff, "dx=", 3) == 0)
	    {
	      for (j = 3; j < stringlength; j++)
		token[j - 3] = buff[j];

	      sscanf (token, "%lf", dx);
	    }

	  if (strncmp (buff, "dy=", 3) == 0)
	    {
	      for (j = 3; j < stringlength; j++)
		token[j - 3] = buff[j];

	      sscanf (token, "%lf", dy);
	    }

	  if (strncmp (buff, "dz=", 3) == 0)
	    {
	      for (j = 3; j < stringlength; j++)
		token[j - 3] = buff[j];

	      sscanf (token, "%lf", dz);
	    }

	  if (strncmp (buff, "output=", 7) == 0)
	    {
	      for (j = 7; j < stringlength; j++)
		token[j - 7] = buff[j];

	      if (filename == NULL)
		{
		  filename = (char *) calloc (stringlength, sizeof (char));
		}
	      sscanf (token, "%s", filename);
	    }

	  if (strncmp (buff, "celltype=", 9) == 0)
	    {
	      for (j = 9; j < stringlength; j++)
		token[j - 9] = buff[j];

	      sscanf (token, "%i", celltype);
	    }

	  if (strncmp (buff, "help", 4) == 0)
	    {
	      print_help ();
	      exit (0);
	    }
	  free (buff);
	  free (token);
	}
      else if (strncmp (argv[i], "-", 1) == 0)
	{
	  for (j = 1; j < stringlength; j++)
	    {
	      if (argv[i][j] == 'z')
		{
		  *zelev = 1;
		}
	      if (argv[i][j] == 'p')
		{
		  *pdata = 1;
		}
	      if (argv[i][j] == 'q')
		{
		  quite = 1;
		}
	      if (argv[i][j] == 'h')
		{
		  print_help ();
		  exit (0);
		}
	    }
	}
      else if ((strncmp (argv[i], "help", 4) == 0))
	{
	  print_help ();
	  exit (0);
	}
    }

  /*summarize */
  if (quite == 0)
    {
      printf ("Cell type %i\n", *celltype);
      printf ("Number of cells in x-direction: %i\n", *cellx);
      printf ("Number of cells in y-direction: %i\n", *celly);
      printf ("Distance between points in x-direction: %f\n", *dx);
      printf ("Distance between points in y-direction: %f\n", *dy);
      printf ("Distance between points in z-direction: %f\n", *dz);
      printf ("Filename: %s\n", filename);
      if (*zelev == 1)
	printf ("Enable variable z elevation\n");
      if (*pdata == 1)
	printf ("Enable point data indices\n");
    }

  return filename;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int
get_number_of_points_of_cell (celltype)
{
  switch (celltype)
    {
    case VTK_POLY_VERTEX:
      return 1;
    case VTK_QUAD:
      return 4;
    case VTK_HEXAHEDRON:
      return 8;
    case VTK_QUADRATIC_TRIANGLE:
      return 6;
    case VTK_QUADRATIC_QUAD:
      return 8;
    case VTK_QUADRATIC_LINEAR_QUAD:
      return 6;
    case VTK_BIQUADRATIC_QUAD:
      return 9;
    case VTK_QUADRATIC_WEDGE:
      return 15;
    case VTK_QUADRATIC_LINEAR_WEDGE:
      return 12;
    case VTK_BIQUADRATIC_QUADRATIC_WEDGE:
      return 18;
    case VTK_QUADRATIC_HEXAHEDRON:
      return 20;
    case VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON:
      return 24;
    case VTK_TRIQUADRATIC_HEXAHEDRON:
      return 27;
    case VTK_CUBIC_EDGE:
      return 4;
    case VTK_BICUBIC_QUAD:
      return 16;
    case VTK_TRICUBIC_HEXAHEDRON:
      return 64;
    }
  return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_ugrid_cells (FILE * out, int cellx, int celly, int celltype)
{
  int nodenum = get_number_of_points_of_cell (celltype);
  int i, j;
  int xoff = 0, yoff = 0, zoff = 0;
  int xnum, ynum, znum;
  compute_node_numbers (cellx, celly, celltype, &xnum, &ynum, &znum);

  if (quite == 0)
    {
      printf ("Creating cells\n");
    }

  if (celltype == VTK_QUADRATIC_TRIANGLE ||
      celltype == VTK_QUADRATIC_WEDGE ||
      celltype == VTK_QUADRATIC_LINEAR_WEDGE || celltype == VTK_BIQUADRATIC_QUADRATIC_WEDGE)
    {
      fprintf (out, "CELLS %i %i\n", cellx * celly * 2, 2 * cellx * celly * nodenum + 2 * cellx * celly);
    }
  else if (celltype == VTK_CUBIC_EDGE)
    {
      fprintf (out, "CELLS %i %i\n", cellx * celly * 8, 8 * cellx * celly * nodenum + 8 * cellx * celly);
    }
  else
    {
      fprintf (out, "CELLS %i %i\n", cellx * celly, cellx * celly * nodenum + cellx * celly);
    }

  if (celltype == VTK_POLY_VERTEX)
    {
      for (i = 0; i < cellx * celly; i++)
	{
	  fprintf (out, "%i ", nodenum);
	  for (j = 0; j < nodenum; j++)
	    {
	      fprintf (out, "%i ", i * nodenum + j);
	    }
	  fprintf (out, "\n");
	}
    }
  else
    {
      for (j = 0; j < celly; j++)
	{
	  for (i = 0; i < cellx; i++)
	    {
	      print_single_cell (out, xnum, ynum, znum, &xoff, &yoff, &zoff, celltype);
	    }
	}
    }

  if (celltype == VTK_QUADRATIC_TRIANGLE ||
      celltype == VTK_QUADRATIC_WEDGE ||
      celltype == VTK_QUADRATIC_LINEAR_WEDGE || celltype == VTK_BIQUADRATIC_QUADRATIC_WEDGE)
    {
      fprintf (out, "CELL_TYPES %i \n", cellx * celly * 2);
      for (i = 0; i < cellx * celly * 2; i++)
	fprintf (out, "%i\n", celltype);
    }
  else if (celltype == VTK_CUBIC_EDGE)
    {
      fprintf (out, "CELL_TYPES %i \n", 8 * cellx * celly);
      for (i = 0; i < 8 * cellx * celly; i++)
	fprintf (out, "%i\n", celltype);
    }
  else
    {
      fprintf (out, "CELL_TYPES %i \n", cellx * celly);
      for (i = 0; i < cellx * celly; i++)
	fprintf (out, "%i\n", celltype);
    }
  return;
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
compute_node_numbers (int cellx, int celly, int celltype, int *xnum, int *ynum, int *znum)
{
  switch (celltype)
    {
    case VTK_POLY_VERTEX:
      *xnum = cellx;
      *ynum = celly;
      *znum = 1;
      break;
    case VTK_QUAD:
      *xnum = cellx + 1;
      *ynum = celly + 1;
      *znum = 1;
      break;
    case VTK_HEXAHEDRON:
      *xnum = cellx + 1;
      *ynum = celly + 1;
      *znum = 2;
      break;
    case VTK_QUADRATIC_LINEAR_QUAD:
      *xnum = cellx * 2 + 1;
      *ynum = celly + 1;
      *znum = 1;
    case VTK_QUADRATIC_QUAD:
    case VTK_QUADRATIC_TRIANGLE:
    case VTK_BIQUADRATIC_QUAD:
      *xnum = cellx * 2 + 1;
      *ynum = celly * 2 + 1;
      *znum = 1;
      break;
    case VTK_QUADRATIC_LINEAR_WEDGE:
      *xnum = cellx * 2 + 1;
      *ynum = celly * 2 + 1;
      *znum = 2;
      break;
    case VTK_QUADRATIC_WEDGE:
    case VTK_BIQUADRATIC_QUADRATIC_WEDGE:
    case VTK_TRIQUADRATIC_HEXAHEDRON:
    case VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON:
    case VTK_QUADRATIC_HEXAHEDRON:
      *xnum = cellx * 2 + 1;
      *ynum = celly * 2 + 1;
      *znum = 3;
      break;
    case VTK_CUBIC_EDGE:
    case VTK_BICUBIC_QUAD:
      *xnum = cellx * 3 + 1;
      *ynum = celly * 3 + 1;
      *znum = 1;
      break;
    case VTK_TRICUBIC_HEXAHEDRON:
      *xnum = cellx * 3 + 1;
      *ynum = celly * 3 + 1;
      *znum = 4;
      break;
    default:
      printf ("ERROR:  cell type %i is not supported\n", celltype);
      exit (0);
    }

  return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_ugrid_point_data (FILE * out, int xnum, int ynum, int znum, int pdata)
{
  int i, j, k;
  fprintf (out, "POINT_DATA %i\n", xnum * ynum * znum);
  fprintf (out, "SCALARS scalars float 1\n");
  fprintf (out, "LOOKUP_TABLE default\n");

  if (quite == 0)
    {
      printf ("Creating point data\n");
    }

  for (k = 0; k < znum; k++)
    for (j = 0; j < ynum; j++)
      for (i = 0; i < xnum; i++)
	if (pdata == 1)
	  fprintf (out, "%i ", k * xnum * ynum + j * xnum + i);
	else
	  fprintf (out, "%g ", (double) (k) + sin ((double) i / (double) (0.1 * xnum)) +
		   cos ((double) j / (0.1 * ynum)));

  fprintf (out, "\n");

  return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_ugrid_points (FILE * out, int xnum, int ynum, int znum, double dx, double dy, double dz, int zelev)
{
  int i, j, k;
  double scale = 0;

  if (quite == 0)
    {
      printf ("Creating points\n");
    }

  fprintf (out, "POINTS %i float\n", xnum * ynum * znum);

  if (zelev == 1)
    scale = 1;

  for (k = 0; k < znum; k++)
    for (j = 0; j < ynum; j++)
      for (i = 0; i < xnum; i++)
	fprintf (out, "%g %g %g\n", dx * (double) (i), dy * (double) (j),
		 dz * ((double) (k) + scale * sin ((double) i / (double) (0.1 * xnum)) +
		       scale * cos ((double) j / (0.1 * ynum))));

  return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_ugrid_header (FILE * out)
{
  fprintf (out, "# vtk DataFile Version 2.0\n");
  fprintf (out, "Created with vtkcellgen from Soeren Gebbert\n");
  fprintf (out, "ASCII\n");
  fprintf (out, "DATASET UNSTRUCTURED_GRID\n");
  return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int
print_single_cell (FILE * out, int cols, int rows, int depths, int *xoffset, int *yoffset, int *zoffset, int celltype)
{
  int retval = 1;
  int nodenum = get_number_of_points_of_cell (celltype);

  switch (celltype)
    {
    case VTK_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_QUADRATIC_TRIANGLE:
      /*first */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      /*second */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_LINEAR_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_QUADRATIC_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_BIQUADRATIC_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_LINEAR_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      /*second */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + cols * rows);

      /*second */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_BIQUADRATIC_QUADRATIC_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + cols * rows);
      /*face center nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      /*second */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      /*face center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*middle quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*middle quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      /*middle side nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 + 1 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_TRIQUADRATIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*middle quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      /*mid side nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 + 1 * cols * rows);
      /*vertical mid side nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 1 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_CUBIC_EDGE:
       /*I*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset) + *xoffset + 2);
       /*II*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 2);
       /*III*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 2);
       /*IV*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 3) + *xoffset + 2);
      /*X-direction */
       /*I*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset);
       /*II*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1);
       /*III*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 2);
       /*IV*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 3);

      *xoffset += 3;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 3;
	}
      break;
    case VTK_BICUBIC_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);

      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1);

      *xoffset += 3;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 3;
	}
      break;
    case VTK_TRICUBIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 * cols * rows);

      /*lower midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);

      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 * cols * rows);

      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 * cols * rows);

      /*face I */
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);

      /*face II */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 + 2 * cols * rows);

      /*face III */
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 1 + 2 * cols * rows);

      /*face IV */
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 + 2 * cols * rows);

      /*face V */
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 0 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 0 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 0 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 0 * cols * rows);

      /*face VI */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 3 * cols * rows);

      /*volume center nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);

      *xoffset += 3;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 3;
	}
      break;
    }

  return retval;
}

